//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:mirrors';
import 'dart:typed_data';

import 'dynamic_library.dart';
import 'native.dart';
import 'util.dart';
import 'model/field.dart';

TypedData encodeStruct(dynamic object) {
  var totalSize = structSize(object);

  var offset = 0;
  TypedData data = new Uint8List(totalSize);
  ByteData view = new ByteData.view(data.buffer);
  iterateFields(object, (field, type, fieldMirror){
    var size = fieldSize(field, type);
    offset += align(offset, size);
    switch (type){
      case C_CHAR:
      case C_UCHAR:
      case C_SHORT:
      case C_USHORT:
      case C_INT:
      case C_UINT:
      case C_LONG:
      case C_ULONG:
      case C_LONGLONG:
      case C_ULONGLONG:
      case C_POINTER:
        setInt(view, size, offset, field != null ? (field as int) : 0);
        break;
      case C_BOOL:
        setInt(view, size, offset, (field as bool) ? 1 : 0);
        break;
      case C_FLOAT:
      case C_DOUBLE:
        setFloat(view, size, offset, field != null ? (field as double) : 0.0);
        break;
      case C_STRING:
        TypedData data = new Uint8List.fromList((field as String).codeUnits);
        setInt(view, size, offset, getPointer(data));
        break;
      case C_STRUCT:
        TypedData data = encodeStruct(field);
        setInt(view, size, offset, getPointer(data));
        break;
    }

    offset += size;
  });

  //print('encode $data ${data.toList().length}');
  return data;
}

dynamic decodeStruct(TypedData data, dynamic object) {
  //print('decode $data');

  ByteData view = new ByteData.view(data.buffer);
  InstanceMirror classMirror = reflect(object);

  var offset = 0;
  iterateFields(object, (field, type, fieldMirror) {
    var size = fieldSize(field, type);
    offset += align(offset, size);

    dynamic value;
    switch (type){
      case C_CHAR:
      case C_UCHAR:
      case C_SHORT:
      case C_USHORT:
      case C_INT:
      case C_UINT:
      case C_LONG:
      case C_ULONG:
      case C_LONGLONG:
      case C_ULONGLONG:
      case C_POINTER:
        value = getInt(view, size, offset);
        break;
      case C_BOOL:
        value = getInt(view, size, offset);
        break;
      case C_FLOAT:
      case C_DOUBLE:
        value = getFloat(view, size, offset);
        break;
      case C_STRING:
        value = "";
        break;
    }

    classMirror.setField(fieldMirror.simpleName, value);
    offset += size;
  });
  return object;
}

int fieldSize(dynamic field, int type) {
  int size;
  switch (type){
    case C_STRING:
    case C_STRUCT:
      size = DynamicLibrary.ctypeToSize(C_POINTER);
      break;
    default:
      size = DynamicLibrary.ctypeToSize(type);
      break;
  }
  return size;
}

int structSize(dynamic object) {
  var totalSize = 0;
  iterateFields(object, (field, type, fieldMirror){
    var size = fieldSize(field, type);
    totalSize += align(totalSize, size);
    totalSize += size;
  });
  return totalSize;
}


void iterateFields(dynamic struct, dynamic process(dynamic field, int type, DeclarationMirror fieldMirror)){
  ClassMirror classMirror = reflect(struct).type;

  var fields = classMirror.declarations.values.where((d)=>true);

  for (var fieldMirror in fields){
    var fieldTag = findDeclTag(fieldMirror, Field);
    if (fieldTag == null) continue;

    var fieldType = fieldTag != null ? (fieldTag.getField(#type).reflectee as int) : DynamicLibrary.dartTypeToCtype(fieldMirror.runtimeType);
    var field = reflect(struct).getField(fieldMirror.simpleName).reflectee;

    process(field, fieldType, fieldMirror);
  }
}

int align(int offset, int size){
   return offset % size != 0 ? (size - offset % size) : 0;
}

