//  Copyright 2016 root.ext@gmail.com
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import 'dart:mirrors';
import 'dart:typed_data';

import 'coder.dart';
import 'native.dart';
import 'util.dart';
import 'model/library.dart';
import 'model/method.dart';
import 'model/out.dart';
import 'model/param.dart';

typedef int Resolver(DynamicLibrary that, Invocation invocation);

class DynamicLibrary<T> {
  final int STACK_SIZE = 256;
  static List<int> sizes = getSizes();

  Type type;
  int libraryPointer;
  Resolver resolver;

  DynamicLibrary([Resolver resolver]){
    this.type = T;
    assert(type != dynamic);
    //print('type $type');

    this.resolver = resolver;

    var libTag = findClassTag(reflectClass(type), Library);
    assert(libTag != null);
    var name = (libTag.getField(#name).reflectee as String);
    var altName = (libTag.getField(#altName).reflectee as String);
    assert(name != null);

    this.libraryPointer = loadLibrary(name);
    if (this.libraryPointer == 0 && altName != null){
      this.libraryPointer = loadLibrary(altName);
    }
    //FIXME assert(libraryPointer != 0);
    //print('lib pointer $libraryPointer');
  }

  dynamic noSuchMethod(Invocation invocation){
    //print('method ${MirrorSystem.getName(invocation.memberName)}');

    var classMirror = reflectClass(type);
    var methods = classMirror.instanceMembers.values
        .where((e)=> e.isRegularMethod /*&& e.owner == classMirror*/ && e.simpleName == invocation.memberName).toList();
    assert(methods.isNotEmpty);
    var methodMirror = methods.first;

    //Array with TypedData object for params
    var paramsData = new List<TypedData>(invocation.positionalArguments.length);
    var methodPointer = resolveMethod(invocation);

    var vm = createCallVM(STACK_SIZE);
    assert(vm != 0);

    before(invocation, methodMirror, vm, paramsData);
    var result = call(methodMirror, vm, methodPointer);
    after(invocation, methodMirror, paramsData);

    freeCallVM(vm);
    return result;
  }

  int resolveMethod(Invocation invocation){
    if (this.resolver != null){
      var pointer = this.resolver(this, invocation);
      return pointer;
    }

    var methodPointer = findSymbol(this.libraryPointer, MirrorSystem.getName(invocation.memberName));
    assert(methodPointer != 0);
    //print('method pointer $methodPointer');
    return methodPointer;
  }

  void iterateParams(Invocation invocation, MethodMirror methodMirror,
    void processParam(dynamic value, Param Param, ParameterMirror paramMirror, int index)){

    var paramMirrors = methodMirror.parameters;

    int index = 0;
    for (var paramMirror in paramMirrors) {
      //print('param ${MirrorSystem.getName(param.simpleName)}');
      var paramTag = findParamTag(paramMirror, Param);
      var param = paramTag != null ? (paramTag.reflectee as Param) : new Param(dartTypeToCtype(paramMirror.type));
      var value = invocation.positionalArguments[index];

      processParam(value, param, paramMirror, index);
      index++;
    }
  }


  void before(Invocation invocation, MethodMirror methodMirror, int vm, List<TypedData> data) {
    var nativeParams = new List<Object>();
    var signature = new List<String>();

    iterateParams(invocation, methodMirror, (value, param, mirror, index){
      signature.add(new String.fromCharCodes([param.type]));
      nativeParams.add(value);

      switch (param.type){
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          if (param.out){
            data[index] = new Uint8List((value.value as String).length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          } else {
            nativeParams[nativeParams.length-1] = new Uint8List.fromList((value as String).codeUnits);
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTINT: {
          if (param.out){
            data[index] = new Uint32List((value.value as List<int>).length);
            nativeParams[nativeParams.length-1] = data[index];
            signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
          }
        }
        break;
        case LISTSTRING: {
          if (param.out) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          if (param.out) {
            data[index] = encodeStruct(value.value);
          } else {
            data[index] = encodeStruct(value);
          }

          nativeParams[nativeParams.length-1] = data[index];
          signature[signature.length-1] = new String.fromCharCodes([TYPEDDATA]);
        }
        break;
      }
    });
    //print(nativeParams);
    pushParameters(vm, signature.join(''), nativeParams);
  }

  void after(Invocation invocation, MethodMirror methodMirror, List<TypedData> data) {

    iterateParams(invocation, methodMirror, (value, param, mirror, index) {
      switch (param.type) {
        case C_POINTER: {

        }
        break;
        case C_STRING: {
          if (param.out) {
            Uint8List list = (data[index] as Uint8List);
            value.value = new String.fromCharCodes(list);
          }
        }
        break;
        case LISTINT: {
          if (param.out){
            Uint32List list = (data[index] as Uint32List);
            value.value = list.toList();
          }
        }
        break;
        case LISTSTRING: {
          if (param.out) {
            throw new UnimplementedError();
          }
        }
        break;
        case C_STRUCT: {
          if (param.out) {
            value.value = decodeStruct(data[index], value.value);
          }
        }
        break;
      }
    });
  }

  dynamic call(MethodMirror methodMirror, int vm, int methodPointer){

    var retTag = findMethodTag(methodMirror, Method);
    var retType = retTag != null ? (retTag.getField(#type).reflectee  as int) : dartTypeToCtype(methodMirror.returnType);

    var result;
    if (retType == VOID){
      callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    } else {
      result = callFunction(vm, methodPointer, new String.fromCharCodes([retType]));
    }

    return result;
  }

  static int dartTypeToCtype(Object type){
    if (type == reflectClass(int)){
      return C_INT;
    } else if (type == reflectClass(double)){
      return C_DOUBLE;
    } else if (type == reflectClass(bool)){
      return C_BOOL;
    } else if (type == reflectClass(String)){
      return C_STRING;
    } else {
      return VOID;
    }
  }

  static int ctypeToSize(int type){
    switch (type){
      case C_CHAR:
      case C_UCHAR:
        return sizes[0];
      case C_SHORT:
      case C_USHORT:
        return sizes[1];
      case C_INT:
      case C_UINT:
        return sizes[2];
      case C_LONG:
      case C_ULONG:
        return sizes[3];
      case C_LONGLONG:
      case C_ULONGLONG:
        return sizes[4];
      case C_BOOL:
        return sizes[5];
      case C_FLOAT:
        return sizes[6];
      case C_DOUBLE:
        return sizes[7];
      case C_POINTER:
        return sizes[8];
      case C_STRING:
        return sizes[8];
      default:
        return 0;
    }
  }
}


